/* Services -- main source file.
 * Copyright (c) 1996-1998 Andy Church <achurch@dragonfire.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "services.h"
#include "timeout.h"


/******** Global variables! ********/

/* These can all be set by options. */
char *remote_server = REMOTE_SERVER;	/* -remote server[:port] */
int remote_port = REMOTE_PORT;
char *server_name = SERVER_NAME;	/* -name servername */
char *server_desc = SERVER_DESC;	/* -desc serverdesc */
char *services_user = SERVICES_USER;	/* -user username */
char *services_host = SERVICES_HOST;	/* -host hostname */
char *services_dir = SERVICES_DIR;	/* -dir directory */
char *log_filename = LOG_FILENAME;	/* -log filename */
char *time_zone = TIMEZONE;		/* -tz timezone */
int update_timeout = UPDATE_TIMEOUT;	/* -update secs */
int expire_timeout = EXPIRE_TIMEOUT;	/* -expire secs */
int debug = 0;				/* -debug */
int readonly = 0;			/* -readonly */
int nofork = 0;				/* -nofork */

/* Set to 1 if we are to quit */
int quitting = 0;

/* Set to 1 if we are to quit after saving databases */
int delayed_quit = 0;

/* Contains a message as to why services is terminating */
char *quitmsg = NULL;

/* Input buffer - global, so we can dump it if something goes wrong */
char inbuf[BUFSIZE];

/* Socket for talking to server */
int servsock = -1;

/* Should we update the databases now? */
int save_data = 0;

/* At what time were we started? */
time_t start_time;


/******** Local variables! ********/

/* Set to 1 if we are waiting for input */
static int waiting = 0;

/* Set to 1 after we've set everything up */
static int started = 0;

/* If we get a signal, use this to jump out of the main loop. */
static jmp_buf panic_jmp;

/*************************************************************************/

/* If we get a weird signal, come here. */

static void sighandler(int signum)
{
    if (started) {
	if (signum == SIGHUP) {  /* SIGHUP = save databases and restart */
	    save_data = -2;
	    signal(SIGHUP, SIG_IGN);
	    return;
	} else if (signum == SIGTERM || signum == SIGINT || signum == SIGQUIT) {
	    /* nothing */
	} else if (!waiting) {
	    log("PANIC! buffer = %s", inbuf);
	    /* Cut off if this would make IRC command >510 characters. */
	    if (strlen(inbuf) > 448) {
		inbuf[446] = '>';
		inbuf[447] = '>';
		inbuf[448] = 0;
	    }
	    wallops(NULL, "PANIC! buffer = %s\r\n", inbuf);
	} else if (waiting < 0) {
	    if (waiting == -1) {
		log("PANIC! in timed_update (%s)", strsignal(signum));
		wallops(NULL, "PANIC! in timed_update (%s)", strsignal(signum));
	    } else {
		log("PANIC! waiting=%d (%s)", waiting, strsignal(signum));
		wallops(NULL, "PANIC! waiting=%d (%s)", waiting, strsignal(signum));
	    }
	}
    }
    if (signum == SIGUSR1 || !(quitmsg = malloc(BUFSIZE))) {
	quitmsg = "Out of memory!";
	quitting = 1;
    } else {
#if HAVE_STRSIGNAL
	snprintf(quitmsg, BUFSIZE, "Services terminating: %s", strsignal(signum));
#else
	snprintf(quitmsg, BUFSIZE, "Services terminating on signal %d", signum);
#endif
	quitting = 1;
    }
    if (started)
	longjmp(panic_jmp, 1);
    else {
	log("%s", quitmsg);
	exit(1);
    }
}

/*************************************************************************/
/*************************************************************************/

static FILE *logfile;

/*************************************************************************/

/* Open the log file. */

void open_log(void)
{
    if (logfile)
	return;
    logfile = fopen(log_filename, "a");
    if (logfile)
	setbuf(logfile, NULL);
}

/* Close the log file. */

void close_log(void)
{
    if (!logfile)
	return;
    fclose(logfile);
    logfile = NULL;
}

/*************************************************************************/

/* Log stuff to stderr with a datestamp. */

void log(const char *fmt, ...)
{
    va_list args;
    time_t t;
    struct tm tm;
    char buf[256];
    int errno_save = errno;

    va_start(args, fmt);
    time(&t);
    tm = *localtime(&t);
#if HAVE_GETTIMEOFDAY
    if (debug) {
	char *s;
	struct timeval tv;
	struct timezone tz;
	gettimeofday(&tv, &tz);
	strftime(buf, sizeof(buf)-1, "[%b %d %H:%M:%S", &tm);
	s = buf + strlen(buf);
	s += snprintf(s, sizeof(buf)-(s-buf), ".%06d", tv.tv_usec);
	strftime(s, sizeof(buf)-(s-buf)-1, " %Y] ", &tm);
    } else {
#endif
	strftime(buf, sizeof(buf)-1, "[%b %d %H:%M:%S %Y] ", &tm);
#if HAVE_GETTIMEOFDAY
    }
#endif
    if (logfile) {
	fputs(buf, logfile);
	vfprintf(logfile, fmt, args);
	fputc('\n', logfile);
    }
    if (nofork) {
	fputs(buf, stderr);
	vfprintf(stderr, fmt, args);
	fputc('\n', stderr);
    }
    errno = errno_save;
}

/* Like log(), but tack a ": " and a system error message (as returned by
 * strerror()) onto the end.
 */

void log_perror(const char *fmt, ...)
{
    va_list args;
    time_t t;
    struct tm tm;
    char buf[256];
    int errno_save = errno;

    va_start(args, fmt);
    time(&t);
    tm = *localtime(&t);
#if HAVE_GETTIMEOFDAY
    if (debug) {
	char *s;
	struct timeval tv;
	struct timezone tz;
	gettimeofday(&tv, &tz);
	strftime(buf, sizeof(buf)-1, "[%b %d %H:%M:%S", &tm);
	s = buf + strlen(buf);
	s += snprintf(s, sizeof(buf)-(s-buf), ".%06d", tv.tv_usec);
	strftime(s, sizeof(buf)-(s-buf)-1, " %Y] ", &tm);
    } else {
#endif
	strftime(buf, sizeof(buf)-1, "[%b %d %H:%M:%S %Y] ", &tm);
#if HAVE_GETTIMEOFDAY
    }
#endif
    if (logfile) {
	fputs(buf, logfile);
	vfprintf(logfile, fmt, args);
	fprintf(logfile, ": %s\n", strerror(errno_save));
    }
    if (nofork) {
	fputs(buf, stderr);
	vfprintf(stderr, fmt, args);
	fprintf(stderr, ": %s\n", strerror(errno_save));
    }
    errno = errno_save;
}

/*************************************************************************/

/* We've hit something we can't recover from.  Let people know what
 * happened, then go down.
 */

void fatal(const char *fmt, ...)
{
    va_list args;
    time_t t;
    struct tm tm;
    char buf[256], buf2[4096];

    va_start(args, fmt);
    time(&t);
    tm = *localtime(&t);
#if HAVE_GETTIMEOFDAY
    if (debug) {
	char *s;
	struct timeval tv;
	struct timezone tz;
	gettimeofday(&tv, &tz);
	strftime(buf, sizeof(buf)-1, "[%b %d %H:%M:%S", &tm);
	s = buf + strlen(buf);
	s += snprintf(s, sizeof(buf)-(s-buf), ".%06d", tv.tv_usec);
	strftime(s, sizeof(buf)-(s-buf)-1, " %Y] ", &tm);
    } else {
#endif
	strftime(buf, sizeof(buf)-1, "[%b %d %H:%M:%S %Y] ", &tm);
#if HAVE_GETTIMEOFDAY
    }
#endif
    vsnprintf(buf2, sizeof(buf2), fmt, args);
    if (logfile)
	fprintf(logfile, "%s FATAL: %s\n", buf, buf2);
    if (nofork)
	fprintf(stderr, "%s FATAL: %s\n", buf, buf2);
    if (servsock >= 0)
	wallops(NULL, "FATAL ERROR!  %s", buf2);
    exit(1);
}

/* Same thing, but do it like perror().
 */

void fatal_perror(const char *fmt, ...)
{
    va_list args;
    time_t t;
    struct tm tm;
    char buf[256], buf2[4096];
    int errno_save = errno;

    va_start(args, fmt);
    time(&t);
    tm = *localtime(&t);
#if HAVE_GETTIMEOFDAY
    if (debug) {
	char *s;
	struct timeval tv;
	struct timezone tz;
	gettimeofday(&tv, &tz);
	strftime(buf, sizeof(buf)-1, "[%b %d %H:%M:%S", &tm);
	s = buf + strlen(buf);
	s += snprintf(s, sizeof(buf)-(s-buf), ".%06d", tv.tv_usec);
	strftime(s, sizeof(buf)-(s-buf)-1, " %Y] ", &tm);
    } else {
#endif
	strftime(buf, sizeof(buf)-1, "[%b %d %H:%M:%S %Y] ", &tm);
#if HAVE_GETTIMEOFDAY
    }
#endif
    vsnprintf(buf2, sizeof(buf2), fmt, args);
    if (logfile)
	fprintf(logfile, "%s FATAL: %s: %s\n", buf, buf2, strerror(errno_save));
    if (stderr)
	fprintf(stderr, "%s FATAL: %s: %s\n", buf, buf2, strerror(errno_save));
    if (servsock >= 0)
	wallops(NULL, "FATAL ERROR!  %s: %s", buf2, strerror(errno_save));
    exit(1);
}

/*************************************************************************/
/*************************************************************************/

/* Send a NICK command for the given pseudo-client.  If `user' is NULL,
 * send NICK commands for all the pseudo-clients. */

#if defined(IRC_DALNET)
# ifdef IRC_DAL4_4_15
#  define NICK(nick,name) \
    do { \
	send_cmd(NULL, "NICK %s 1 %lu %s %s %s 0 :%s", (nick), time(NULL), \
		services_user, services_host, server_name, (name)); \
    } while (0)
# else
#  define NICK(nick,name) \
    do { \
	send_cmd(NULL, "NICK %s 1 %lu %s %s %s :%s", (nick), time(NULL), \
		services_user, services_host, server_name, (name)); \
    } while (0)
# endif
#elif defined(IRC_UNDERNET)
# define NICK(nick,name) \
    do { \
	send_cmd(server_name, "NICK %s 1 %lu %s %s %s :%s", (nick), time(NULL),\
		services_user, services_host, server_name, (name)); \
    } while (0)
#elif defined(IRC_TS8)
# define NICK(nick,name) \
    do { \
	send_cmd(NULL, "NICK %s :1", (nick)); \
	send_cmd((nick), "USER %ld %s %s %s :%s", time(NULL), \
		services_user, services_host, server_name, (name)); \
    } while (0)
#else
# define NICK(nick,name) \
    do { \
	send_cmd(NULL, "NICK %s :1", (nick)); \
	send_cmd((nick), "USER %ld %s %s %s :%s", \
		services_user, services_host, server_name, (name)); \
    } while (0)
#endif

void introduce_user(const char *user)
{
    if (!user || stricmp(user, s_NickServ) == 0) {
	NICK(s_NickServ, "Nickname Server");
#ifndef IRC_DALNET
	send_cmd(s_NickServ, "MODE %s +o", s_NickServ);
#endif
    }
    if (!user || stricmp(user, s_ChanServ) == 0) {
	NICK(s_ChanServ, "Channel Server");
#ifndef IRC_DALNET
	send_cmd(s_ChanServ, "MODE %s +o", s_ChanServ);
#endif
    }
    if (!user || stricmp(user, s_HelpServ) == 0) {
	NICK(s_HelpServ, "Help Server");
    }
    if (!user || stricmp(user, IRCIIHELP_NAME) == 0) {
	NICK(IRCIIHELP_NAME, "ircII Help Server");
    }
    if (!user || stricmp(user, s_MemoServ) == 0) {
	NICK(s_MemoServ, "Memo Server");
#ifndef IRC_DALNET
	send_cmd(s_MemoServ, "MODE %s +o", s_MemoServ);
#endif
    }
    if (!user || stricmp(user, s_OperServ) == 0) {
	NICK(s_OperServ, "Operator Server");
#ifndef IRC_DALNET
	send_cmd(s_OperServ, "MODE %s +oi", s_OperServ);
#else
	send_cmd(s_OperServ, "MODE %s +i", s_OperServ);
#endif
    }
#ifdef DEVNULL_NAME
    if (!user || stricmp(user, DEVNULL_NAME) == 0) {
	NICK(DEVNULL_NAME, "/dev/null -- message sink");
	send_cmd(DEVNULL_NAME, "MODE %s +i", DEVNULL_NAME);
    }
#endif
    if (!user || stricmp(user, s_GlobalNoticer) == 0) {
	NICK(s_GlobalNoticer, "Global Noticer");
	send_cmd(s_GlobalNoticer, "MODE %s +io", s_GlobalNoticer);
    }
}

#undef NICK

/*************************************************************************/
/*************************************************************************/

/* Initialization/cleanup routines: */

/*************************************************************************/

/* Set group name if necessary.  Return 0 if successful (or if RUNGROUP not
 * defined), else print an error message to logfile and return -1.
 */

static int set_group(void)
{
#ifdef RUNGROUP
    struct group *gr;

    setgrent();
    while (gr = getgrent()) {
	if (strcmp(gr->gr_name, RUNGROUP) == 0)
	    break;
    }
    endgrent();
    if (gr) {
	setgid(gr->gr_gid);
	return 0;
    } else {
	fprintf(logfile, "Unknown group `%s'\n", RUNGROUP);
	return -1;
    }
#else
    return 0;
#endif
}

/*************************************************************************/

/* Parse command-line options.  Return 0 if all went well, -1 for an error
 * with an option, or 1 for -help.
 */

static int parse_options(int ac, char **av)
{
    int i;
    char *s, *t;

    for (i = 1; i < ac; i++) {
	s = av[i];
	if (*s == '-') {
	    s++;
	    if (strcmp(s, "remote") == 0) {
		if (++i >= ac) {
		    fprintf(stderr, "-remote requires hostname[:port]\n");
		    return -1;
		}
		s = av[i];
		if (t = strchr(s, ':')) {
		    *t++ = 0;
		    if (atoi(t) > 0)
			remote_port = atoi(t);
		    else {
			fprintf(stderr, "-remote: port number must be a positive integer.  Using default.");
			return -1;
		    }
		}
		remote_server = s;
	    } else if (strcmp(s, "name") == 0) {
		if (++i >= ac) {
		    fprintf(stderr, "-name requires a parameter\n");
		    return -1;
		}
		server_name = av[i];
	    } else if (strcmp(s, "desc") == 0) {
		if (++i >= ac) {
		    fprintf(stderr, "-desc requires a parameter\n");
		    return -1;
		}
		server_desc = av[i];
	    } else if (strcmp(s, "user") == 0) {
		if (++i >= ac) {
		    fprintf(stderr, "-user requires a parameter\n");
		    return -1;
		}
		services_user = av[i];
	    } else if (strcmp(s, "host") == 0) {
		if (++i >= ac) {
		    fprintf(stderr, "-host requires a parameter\n");
		    return -1;
		}
		services_host = av[i];
	    } else if (strcmp(s, "dir") == 0) {
		if (++i >= ac) {
		    fprintf(stderr, "-dir requires a parameter\n");
		    return -1;
		}
		services_dir = av[i];
	    } else if (strcmp(s, "log") == 0) {
		if (++i >= ac) {
		    fprintf(stderr, "-log requires a parameter\n");
		    return -1;
		}
		log_filename = av[i];
	    } else if (strcmp(s, "tz") == 0) {
		if (++i >= ac) {
		    fprintf(stderr, "-tz requires a parameter\n");
		    return -1;
		}
		time_zone = av[i];
	    } else if (strcmp(s, "update") == 0) {
		if (++i >= ac) {
		    fprintf(stderr, "-update requires a parameter\n");
		    return -1;
		}
		s = av[i];
		if (atoi(s) <= 0) {
		    fprintf(stderr, "-update: number of seconds must be positive");
		    return -1;
		} else
		    update_timeout = atol(s);
	    } else if (strcmp(s, "expire") == 0) {
		if (++i >= ac) {
		    fprintf(stderr, "-expire requires a parameter\n");
		    return -1;
		}
		s = av[i];
		if (atoi(s) <= 0) {
		    fprintf(stderr, "-expire: number of seconds must be positive");
		    return -1;
		} else
		    expire_timeout = atol(s);
	    } else if (strcmp(s, "debug") == 0) {
		debug++;
	    } else if (strcmp(s, "readonly") == 0) {
		readonly = 1;
	    } else if (strcmp(s, "nofork") == 0) {
		nofork = 1;
	    } else {
		fprintf(stderr, "Unknown option -%s\n", s);
		return -1;
	    }
	} else {
	    fprintf(stderr, "Non-option arguments not allowed\n");
	    return -1;
	}
    }
    return 0;
}

/*************************************************************************/

/* Remove our PID file.  Done at exit. */

static void remove_pidfile(void)
{
    remove(PID_FILE);
}

/*************************************************************************/

/* Create our PID file and write the PID to it. */

static void write_pidfile(void)
{
    FILE *pidfile;

    pidfile = fopen(PID_FILE, "w");
    if (pidfile) {
	fprintf(pidfile, "%d\n", getpid());
	fclose(pidfile);
	atexit(remove_pidfile);
    } else {
	log_perror("Warning: cannot write to PID file %s", PID_FILE);
    }
}

/*************************************************************************/

/* Overall initialization routine. */

static int init(int ac, char **av)
{
    int i;

    /* Set file creation mask and group ID. */
#ifdef DEFUMASK
    umask(DEFUMASK);
#endif
    if (set_group() < 0)
	return 1;
    
    /* Parse options. */
    parse_options(ac, av);

    /* Chdir to Services data directory. */
    if (chdir(services_dir) < 0) {
	fprintf(stderr, "chdir(%s): %s\n", services_dir, strerror(errno));
	return 1;
    }

    /* Detach ourselves if requested. */
    if (!nofork) {
	if ((i = fork()) < 0) {
	    perror("fork()");
	    return 1;
	} else if (i != 0) {
	    exit(0);
	}
	close(0);
	close(1);
	close(2);
	if (setpgid(0, 0) < 0) {
	    perror("setpgid()");
	    return 1;
	}
    }

    /* Open logfile. */
    open_log();

    /* Write our PID to the PID file. */
    write_pidfile();

    /* Announce ourselves to the logfile. */
#ifdef SKELETON
    log("Services (skeleton version) starting up");
#else
    if (debug || readonly) {
	log("Services starting up (options:%s%s)",
		debug ? " debug" : "", readonly ? " readonly" : "");
    } else {
	log("Services starting up (normal mode)");
    }
#endif
    start_time = time(NULL);

    /* If in read-only mode, close the logfile again. */
    if (readonly)
	close_log();

    /* Set signal handlers. */
    signal(SIGINT, sighandler);
    signal(SIGTERM, sighandler);
    signal(SIGPIPE, SIG_IGN);		/* We don't care about broken pipes */
    signal(SIGQUIT, sighandler);
    signal(SIGSEGV, sighandler);
    signal(SIGBUS, sighandler);
    signal(SIGQUIT, sighandler);
    signal(SIGHUP, sighandler);
    signal(SIGILL, sighandler);
    signal(SIGTRAP, sighandler);
    signal(SIGIOT, sighandler);
    signal(SIGFPE, sighandler);
    signal(SIGALRM, SIG_IGN);		/* Used by sgets() for read timeout */
    signal(SIGUSR2, SIG_IGN);
    signal(SIGCHLD, SIG_IGN);
    signal(SIGWINCH, SIG_IGN);
    signal(SIGTTIN, SIG_IGN);
    signal(SIGTTOU, SIG_IGN);
    signal(SIGTSTP, SIG_IGN);

    signal(SIGUSR1, sighandler);  /* This is our "out-of-memory" panic switch */

    /* Load up databases */
#ifndef SKELETON
    load_ns_dbase();
    if (debug)
	log("debug: Loaded %s database (1/5)", s_NickServ);
    load_cs_dbase();
    if (debug)
	log("debug: Loaded %s database (2/5)", s_ChanServ);
    load_ms_dbase();
    if (debug)
	log("debug: Loaded %s database (3/5)", s_MemoServ);
#endif
    load_os_dbase();
    if (debug)
	log("debug: Loaded %s database (4/5)", s_OperServ);
    load_akill();
    if (debug)
	log("debug: Loaded AKILL database (5/5)");
    log("Databases loaded");

    /* Connect to the remote server */
    servsock = conn(remote_server, remote_port);
    if (servsock < 0)
	fatal_perror("Can't connect to server");
    send_cmd(NULL, "PASS :%s", PASSWORD);
    send_cmd(NULL, "SERVER %s 1 :%s", server_name, server_desc);
    sgets2(inbuf, sizeof(inbuf), servsock);
    if (strnicmp(inbuf, "ERROR", 5) == 0) {
	log("Remote server returned: %s", inbuf);
	return 1;
    }

    /* Bring in our pseudo-clients */
    introduce_user(NULL);

    /* Success! */
    return 0;
}

/*************************************************************************/
/*************************************************************************/

/* Main routine.  (What does it look like? :-) ) */

int main(int ac, char **av, char **envp)
{
    volatile time_t last_update; /* When did we last update the databases? */
    volatile time_t last_expire; /* When did we last expire nicks/channels? */
    volatile time_t last_check;  /* When did we last check timeouts? */
    int i;
    char *progname;


#ifndef SKELETON
    /* Find program name. */
    if (progname = strrchr(av[0], '/'))
	progname++;
    else
	progname = av[0];

    /* Were we run under "listnicks" or "listchans"?  Do appropriate stuff
     * if so. */
    if (strcmp(progname, "listnicks") == 0) {
	do_listnicks(ac, av);
	return 0;
    } else if (strcmp(progname, "listchans") == 0) {
	do_listchans(ac, av);
	return 0;
    }
#endif


    /* Initialization stuff. */
    if ((i = init(ac, av)) != 0)
	return i;


    /* We have a line left over from earlier, so process it first. */
    process();

    last_update = time(NULL);
    last_expire = time(NULL);
    last_check  = time(NULL);

    /* The signal handler routine will drop back here with quitting != 0
     * if it gets called. */
    setjmp(panic_jmp);
    started = 1;


    /*** Main loop. ***/

    while (!quitting) {
	time_t t = time(NULL);

	if (debug >= 2)
	    log("debug: Top of main loop");
	waiting = -3;
	if (!readonly && (save_data || t-last_expire >= expire_timeout)) {
	    if (debug)
		log("debug: Running expire routines");
#ifndef SKELETON
	    waiting = -21;
	    expire_nicks();
	    waiting = -22;
	    expire_chans();
#endif
	    waiting = -25;
	    expire_akills();
	    last_expire = t;
	}
	if (!readonly && (save_data || t-last_update >= update_timeout)) {
	    waiting = -2;
	    if (debug)
		log("debug: Saving databases");
#ifndef SKELETON
	    waiting = -11;
	    save_ns_dbase();
	    waiting = -12;
	    save_cs_dbase();
	    waiting = -13;
	    save_ms_dbase();
#endif
	    waiting = -14;
	    save_os_dbase();
	    waiting = -15;
	    save_akill();
	    if (save_data < 0)
		break;	/* out of main loop */

	    save_data = 0;
	    last_update = t;
	}
	if (delayed_quit)
	    break;
	waiting = -1;
	if (t-last_check >= TIMEOUT_CHECK) {
	    check_timeouts();
	    last_check = t;
	}
	waiting = 1;
	i = (int)sgets2(inbuf, sizeof(inbuf), servsock);
	waiting = 0;
	if (i > 0) {
	    process();
	} else if (i == 0) {
	    if (quitmsg = malloc(BUFSIZE))
		snprintf(quitmsg, BUFSIZE,
			"Read error from server: %s", strerror(errno));
	    else
		quitmsg = "Read error from server";
	    quitting = 1;
	}
	waiting = -4;
    }


    /* Check for restart instead of exit */
    if (save_data == -2) {
#ifdef SERVICES_BIN
	log("Restarting");
	if (!quitmsg)
	    quitmsg = "Restarting";
	send_cmd(server_name, "SQUIT %s :%s", server_name, quitmsg);
	disconn(servsock);
	execve(SERVICES_BIN, av, envp);
	log_perror("Restart failed");
	return 1;
#else
	quitmsg = "Restart attempt failed--SERVICES_BIN not defined (rerun configure)";
#endif
    }

    /* Disconnect and exit */
    if (!quitmsg)
	quitmsg = "Terminating, reason unknown";
    log("%s", quitmsg);
    if (started)
	send_cmd(server_name, "SQUIT %s :%s", server_name, quitmsg);
    disconn(servsock);
    return 0;
}
